import itertools
from datetime import datetime, date, timedelta
from peewee import *
from playhouse.sqlite_ext import FTS5Model, SearchField, RowIDField
from playhouse.flask_utils import FlaskDB
#import sqlite3

# Defer init in app creation
db_wrapper = FlaskDB()

SYNC_STARTED = "S"
SYNC_ERROR = "E"
SYNC_OK = "K"

TAG_GENRE = "G"
TAG_KEYWORD = "K"

class SyncLog(db_wrapper.Model):
    timestamp = TimestampField(utc=True)
    status = CharField(default=SYNC_STARTED)
    description = TextField(default="")

    def __str__(self):
        return f"[{self.status} {self.timestamp}] {self.description}"


def get_last_log():
    """
    Get last successful sync, if any
    """
    return SyncLog.select().where(SyncLog.status == "K").order_by(SyncLog.timestamp.desc()).get_or_none()


class Series(db_wrapper.Model):
    tmdb_id = IntegerField(unique=True)
    imdb_id = CharField(default="")
    name = CharField()
    sort_name = CharField()
    tagline = CharField(default="")
    slug = CharField()
    overview = TextField()
    network = CharField(default="")
    poster_url = CharField(default="")
    fanart_url = CharField(default="")
    vote_average = FloatField(default=0)
    popularity = FloatField(default=0)
    status = FixedCharField(max_length=1)
    language = CharField(max_length=2)
    last_updated_on = DateTimeField(default=datetime.utcnow)

    @property
    def imdb_url(self):
        if self.imdb_id:
            return f"https://www.imdb.com/title/{self.imdb_id}/"
        else:
            return ""

    @property
    def tvdb_url(self):
        return f"https://thetvdb.com/series/{self.slug}"

    def __str__(self):
        return self.name


class SeriesIndex(FTS5Model):
    """
    Full-text search index for series
    """
    rowid = RowIDField()
    name = SearchField()
    content = SearchField() # Contains all other relevant fields

    class Meta:
        database = db_wrapper.database
        options = {'tokenize': 'porter'}


class Tag(db_wrapper.Model):
    slug = CharField()
    name = CharField()
    type = FixedCharField(max_length=1)    

    def __str__(self):
        return self.name


class SeriesTag(db_wrapper.Model):
    series = ForeignKeyField(Series, on_delete="CASCADE")
    tag = ForeignKeyField(Tag, on_delete="CASCADE")

    class Meta:
        primary_key = CompositeKey('series', 'tag')


class Episode(db_wrapper.Model):
    tmdb_id = IntegerField()
    series = ForeignKeyField(Series, backref='episodes', on_delete="CASCADE")
    name = CharField()
    season = SmallIntegerField()
    number = SmallIntegerField()
    aired_on = DateField(null=True)
    overview = TextField(default="")
    last_updated_on = DateTimeField(default=datetime.utcnow)
    thumbnail_url = CharField(default="")

    @property
    def season_episode_id(self):
        return "S{:02}.E{:02}".format(self.season, self.number)

    def __str__(self):
        return f"{self.season_episode_id} '{self.name}'"

    class Meta:
        indexes = (
            (('series', 'season', 'number'), True),
        )


# class EpisodeIndex(FTS5Model):
#     """
#     Full-text search index for episodes
#     """
#     rowid = RowIDField()
#     name = SearchField()
#     overview = SearchField()

#     class Meta:
#         database = db_wrapper.database
#         options = {'tokenize': 'porter'}


class Release(db_wrapper.Model):
    # Enough for BitTorrent 2 SHA-256 hashes
    info_hash = CharField(unique=True, max_length=64)
    # We could change an episode id, so update this FK accordingly
    episode = ForeignKeyField(
        Episode, backref='releases', on_update="CASCADE", on_delete="CASCADE")
    added_on = DateTimeField()
    last_updated_on = DateTimeField()
    size = BigIntegerField()
    magnet_uri = TextField()
    seeders = IntegerField()
    leechers = IntegerField()
    completed = IntegerField()
    name = CharField()
    resolution = SmallIntegerField()

    def __str__(self):
        return self.name


###########
# DB SETUP
###########

def setup():
    db_wrapper.database.create_tables([
        Series,
        SeriesIndex,
        Episode,
        #EpisodeIndex,
        Release,
        Tag,
        SeriesTag,
        SyncLog,
    ], safe=True)


'''
Custom Jinja Filters
'''
import itertools
from videobox import languages

MIN_SEEDERS = 1

def human_date(value):
    return value.strftime("%b %d, %Y")

def human_date_time(value):
    return value.strftime("%b %d, %Y at %H:%M")

def islice(iterable, stop):
    return itertools.islice(iterable, stop)

def networks(value):
    pieces = value.split(", ")
    # Show up to 3 networks
    if len(pieces) < 4:
        return " • ".join(pieces)
    else:
        return f"{pieces[0]} and others"

def torrent_health(value):
    color = "hi"
    if MIN_SEEDERS <= value < 4:
        color = "low"
    elif 4 <= value < 7:
        color = "medium"
    return f'<span class="torrent-{color}">{value}</span>'

def lang(code):
    try:
        return languages.LANGUAGES[code]
    except KeyError:
        return ''

FILTERS = [
    human_date,
    human_date_time,
    torrent_health,
    networks,
    lang,
    islice
]

def init_app(app):
    for func in FILTERS:
        app.add_template_filter(func)



from datetime import datetime, date, timezone, timedelta
from pathlib import Path
from operator import attrgetter
from itertools import groupby, islice
import queue
import flask
from flask import current_app as app
from flask import Flask
from peewee import fn, JOIN
from playhouse.flask_utils import PaginatedQuery, get_object_or_404
import videobox
import videobox.models as models
from videobox.models import Series, Episode, Release, Tag, SeriesTag
import videobox.utilities as utilities
import videobox.sync as sync
from . import bp
from . import queries

MAX_TOP_TAGS = 10
MAX_SEASONS = 2
MIN_SEEDERS = 1
SERIES_PER_PAGE = 6 * 10
APP_DIR = Path.home().joinpath(".videobox")
RESOLUTION_OPTIONS = {
    0: "Any",
    480: "480p",
    720: "720p (HD)",
    1080: "1080p (Full HD)",
    2160: "2160p (4K)",
}
SIZE_OPTIONS = {
    "": "Any",
    "asc": "Smallest",
    "desc": "Largest",
}


@bp.context_processor
def inject_template_vars():
    return {
        "last_sync": models.get_last_log(),
        "version": videobox.__version__
    }


@bp.errorhandler(404)
def page_not_found(e):
    return flask.render_template('404.html'), 404


# ---------
# Home
# ---------

last_server_alert = ''

@bp.route('/')
def home():
    total_series, total_episodes, total_releases = queries.get_library_stats()
    if total_series:        
        today_series = queries.get_today_series()
        today_series_tags = None
        exclude_ids = []
        if today_series:
            today_series_tags = queries.get_series_tags(today_series) 
            exclude_ids = [today_series.id]
        featured_series = queries.get_featured_series(exclude_ids=exclude_ids, days_interval=2).limit(8)
        top_tags = queries.get_nav_tags(8)
        return flask.render_template("home.html", 
                                    server_alert=last_server_alert,
                                    today_series=today_series,
                                    today_series_tags=today_series_tags,
                                    featured_series=featured_series, 
                                    top_tags=top_tags,
                                    total_series=total_series,
                                    total_episodes=total_episodes,
                                    total_releases=total_releases)
    else:
        return flask.render_template("first-import.html")

# ---------
# Search
# ---------


@bp.route('/search')
def search():
    query = flask.request.args.get("query")
    series_ids = [series.rowid for series in queries.search_series(
        utilities.sanitize_query(query))]
    series = queries.get_series_with_ids(series_ids)
    # @@TODO
    # if len(results) == 1:
    #     # Single match, display series detail page
    #     series = models.get_series(results[0].rowid)
    return flask.render_template("search_results.html", found_series=series, search_query=query)


@bp.route('/suggest')
def suggest():
    query = flask.request.args.get("query")
    sanitized_query = utilities.sanitize_query(query)
    search_suggestions = queries.suggest_series(sanitized_query).limit(10)
    return flask.render_template("_suggest.html", search_suggestions=search_suggestions)


# ---------
# Tag
# ---------

@bp.route('/tag')
def tag():    
    tags_all = queries.get_top_series_for_tags()
    tags_series = groupby(tags_all, key=attrgetter('tag_slug', 'tag_name'))
    return flask.render_template("tags.html", tags_series=tags_series)


@bp.route('/tag/<slug>')
def tag_detail(slug):
    page = flask.request.args.get("page", 1, type=int)
    tag = get_object_or_404(Tag, (Tag.slug == slug))
    query = queries.get_series_for_tag(tag)
    paginated_series = PaginatedQuery(query, paginate_by=SERIES_PER_PAGE, page_var="page", check_bounds=True)
    if page == 1:
        return flask.render_template("tag_detail.html", tag=tag, series=paginated_series, page=1, series_count=query.count())
    else:
        # For async requests
        return flask.render_template("_tag-card-grid.html", tag=tag, series=paginated_series, page=page)

# --------- 
# Series Detail
# ---------


@bp.route('/series/<int:series_id>')
def series_detail(series_id):
    #resolution = flask.request.args.get("resolution", type=int, default=0) or flask.request.cookies.get('resolution', type=int, default=0)
    resolution = flask.request.args.get("resolution", type=int, default=0)
    #size_sorting = flask.request.args.get("size", default="") or flask.request.cookies.get('size', default="")
    size_sorting = flask.request.args.get("size", default="")
    today = date.today()
    series = get_object_or_404(Series, (Series.id == series_id))
    series_subquery = queries.get_series_subquery()
    release_cte = queries.release_cte(resolution, size_sorting)
    if resolution or size_sorting:
        episodes_query = (Episode.select(Episode.id, Episode.name, Episode.season, Episode.number, Episode.aired_on, Release.id, Release.name, Release.magnet_uri, Release.resolution, Release.size, Release.seeders, Release.last_updated_on)
                        .join(Release)
                        .switch(Episode)
                        .join(Series)
                        .join(series_subquery, on=(
                            series_subquery.c.id == Series.id))
                        .join(release_cte, on=(Release.id == release_cte.c.release_id))
                        .where((Episode.series == series.id)
                                # Episodes from last 2 seasons only
                                & (series_subquery.c.max_season - Episode.season < MAX_SEASONS)
                                & (Episode.aired_on != None)
                                )
                        .order_by(Episode.season.desc(), Episode.number)
                        .with_cte(release_cte))
    else:
        # Unfiltered
        episodes_query = (Episode.select(Episode.id, Episode.name, Episode.season, Episode.number, Episode.aired_on, Release.id, Release.name, Release.magnet_uri, Release.resolution, Release.size, Release.seeders, Release.last_updated_on)
                        .join(Release, JOIN.LEFT_OUTER)
                        .switch(Episode)
                        .join(Series)
                        .join(series_subquery, on=(
                            series_subquery.c.id == Series.id))
                        .where((Episode.series == series.id)
                                # Episodes from last 2 seasons only
                                & (series_subquery.c.max_season - Episode.season < MAX_SEASONS)
                                & (Episode.aired_on != None)
                                )
                        .order_by(Episode.season.desc(), Episode.number, Release.seeders.desc()))

    # Group by season number
    seasons_episodes = groupby(episodes_query, key=attrgetter('season'))
    series_tags = queries.get_series_tags(series) 

    response = flask.make_response(flask.render_template("series_detail.html", 
        series=series, 
        series_tags=series_tags, 
        seasons_episodes=seasons_episodes, 
        today=today, 
        resolution=resolution, 
        resolution_options=RESOLUTION_OPTIONS, 
        size=size_sorting, 
        size_options=SIZE_OPTIONS))
    # Remember filters across requests
    if resolution:
        response.set_cookie('resolution', str(resolution))
    if size_sorting:
        response.set_cookie('size', size_sorting)    
    return response


@bp.route('/release/<int:release_id>')
def release_detail(release_id):
    release = get_object_or_404(Release, (Release.id == release_id))
    return flask.render_template("_release_detail.html", release=release)


# ---------
# Update database
# ---------

close_message = object()

class MessageAnnouncer(object):
    """
    Server-sent events in Flask without extra dependencies.
    See https://maxhalford.github.io/blog/flask-sse-no-deps/
    """

    def __init__(self):
        self.listeners = []

    def listen(self):
        q = queue.Queue(maxsize=5)
        self.listeners.append(q)
        return q

    def format_sse(self, data, event=None):
        msg = f'data: {data}\n\n'
        if event:
            msg = f'event: {event}\n{msg}'
        return msg

    def announce(self, msg):
        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(msg)
            except queue.Full:
                del self.listeners[i]


announcer = MessageAnnouncer()
sync_worker = None

@bp.route('/update')
def update():

    @flask.copy_current_request_context
    def on_update_progress(message, percent=0):
        data = flask.render_template(
            "_update-dialog.html", message=message, percent=percent)
        msg = announcer.format_sse(data=data, event='updating')
        announcer.announce(msg)

    @flask.copy_current_request_context
    def on_update_done(message, alert):
        data = flask.render_template(
            "_update-dialog-done.html", message=message, percent=100)
        msg = announcer.format_sse(data=data, event='done')
        announcer.announce(msg)
        announcer.announce(close_message)
        global last_server_alert
        last_server_alert = alert

    global sync_worker 
    if sync_worker and sync_worker.is_alive():
        app.logger.warning("Sync is already running, request ignored")
    else:
        sync_worker = sync.SyncWorker(
            app.config["API_CLIENT_ID"], progress_callback=on_update_progress, done_callback=on_update_done)
        sync_worker.start()
    
    return {}, 200

@bp.route('/update-events')
def update_events():    
    def stream():
        # Returns a queue.Queue
        messages = announcer.listen()
        while True:
            # Blocks until a new message arrives
            msg = messages.get()
            if msg is close_message:
                break
            yield msg
    return flask.Response(stream(), mimetype='text/event-stream')



from peewee import chunked, fn
import time
from datetime import datetime, timedelta, timezone
from threading import Thread
from requests.exceptions import HTTPError, ReadTimeout
from flask import current_app
import videobox.api as api
import videobox.models as models
from videobox.models import Tag, SeriesTag, Series, SeriesIndex, Episode, Release, SyncLog

# @@TODO check 
# SQLite >3.32.0 has a limit of total 32766 max variables (SQLITE_MAX_VARIABLE_NUMBER)
# https://stackoverflow.com/a/64419474 
# https://stackoverflow.com/q/35616602
INSERT_CHUNK_SIZE = 999 // 15   # Series class has the max numbes of fields 
REQUEST_CHUNK_SIZE = 450        # Total URI must be < 4096
TIMEOUT_BEFORE_RETRY = 5        # Seconds


class SyncError(Exception):
    pass


class SyncWorker(Thread):

    def __init__(self, client_id, progress_callback=None, done_callback=None):
        super().__init__(name="Sync worker")
        self.app = current_app._get_current_object()
        self.client_id = client_id
        self.progress_callback = progress_callback
        self.done_callback = done_callback

    def run(self):
        last_log = models.get_last_log()
        start_time = time.time()

        # Manually push the app context to make Flask
        #   logger to work on the separate thread
        with self.app.app_context():
            current_log = SyncLog.create(description="Started import/sync")

            alert = ""
            try:
                if last_log:
                    alert, series_count, episode_count, release_count = self.update_library(
                        last_log)
                else:
                    series_count, episode_count, release_count = self.import_library()
            except SyncError as ex:
                self.update_log(current_log, status=models.SYNC_ERROR, description=str(ex))
                if self.done_callback:
                    self.done_callback(str(ex), alert)
                return

            elapsed_time = time.time()-start_time
            if any([series_count, episode_count, release_count]):
                description = f"Added/updated {series_count} series, {episode_count} episodes, and {release_count} torrents"
            else:
                description = "No updates were found"

            # Mark import/sync successful
            self.update_log(current_log, status=models.SYNC_OK, description=description)

            self.app.logger.info(f"Finished in {elapsed_time:.1f}s: {description}")

            if self.done_callback:
                self.done_callback(description, alert)

    def import_library(self):
        series_count, episode_count, release_count = 0, 0, 0
        instant = datetime.utcnow()

        self.app.logger.info("No local database found, starting full import")
        
        if self.progress_callback:
            self.progress_callback("Importing all tags...", 0)

        json = self.do_json_request(
            lambda: api.get_all_tags(self.client_id), retries=3)
        if json:
            if self.progress_callback:
                self.progress_callback("Saving tags to library...", 12.5)
            self.save_tags(json)

        if self.progress_callback:
            self.progress_callback("Importing all series...", 25)

        json = self.do_json_request(
            lambda: api.get_all_series(self.client_id))
        if json:
            if self.progress_callback:
                self.progress_callback("Saving series to library...", 37.5)
            series_count = self.save_series(json, instant)

        if self.progress_callback:
            self.progress_callback("Importing all series tags...", 50)

        json = self.do_json_request(
            lambda: api.get_all_series_tags(self.client_id))
        if json:
            self.save_series_tags(json)

        if self.progress_callback:
            self.progress_callback("Importing all episodes...", 62.5)

        json = self.do_json_request(
            lambda: api.get_all_episodes(self.client_id))
        if json:
            if self.progress_callback:
                self.progress_callback("Saving episodes to library...", 75)
            episode_count = self.save_episodes(json, instant)

        if self.progress_callback:
            self.progress_callback("Importing all torrents...", 87.5)

        json = self.do_json_request(
            lambda: api.get_all_releases(self.client_id))
        if json:
            if self.progress_callback:
                self.progress_callback("Saving torrents to library...", 87.5)            
            release_count = self.save_releases(json, instant)

        return series_count, episode_count, release_count

    def update_library(self, last_log):
        series_count, episode_count, release_count = 0, 0, 0

        self.app.logger.info("Last update done at {0} UTC, requesting updates since then".format(
            last_log.timestamp.isoformat()))
        if self.progress_callback:
            self.progress_callback("Getting updated series...")
        # Ensure UTC tz
        json = self.do_json_request(lambda: api.get_updated_series(
            self.client_id, last_log.timestamp.replace(tzinfo=timezone.utc)), retries=3)

        # Save alert from server, if any
        alert = json["alert"]

        tag_ids = json['tags']
        if tag_ids:
            self.app.logger.debug(
                "Got {0} tags, starting update".format(len(tag_ids)))
            tags_count = self.sync_tags(tag_ids)        

        # Grab series
        series_ids = json['series']
        if series_ids:
            self.app.logger.debug(
                "Got {0} series, starting update".format(len(series_ids)))
            series_count = self.sync_series(series_ids)

        # Grab episodes
        episode_ids = json['episodes']
        if episode_ids:
            self.app.logger.debug(
                "Got {0} episodes, starting update".format(len(episode_ids)))
            episode_count = self.sync_episodes(episode_ids)

        # Grab releases
        release_ids = json['releases']
        if release_ids:
            self.app.logger.debug(
                "Got {0} releases, starting update".format(len(release_ids)))
            release_count = self.sync_releases(release_ids)

        return alert, series_count, episode_count, release_count

    def save_tags(self, response):
        """
        Insert new tags and attempt to update existing ones
        """
        count = 0
        self.app.logger.debug("Saving tags to database...")
        for batch in chunked(response, INSERT_CHUNK_SIZE):
            count += (Tag.insert_many(batch)
                      # Replace current tag with new data
                      .on_conflict_replace()
                      .as_rowcount()
                      .execute())
        return count

    def sync_tags(self, remote_ids):
        count = 0

        # Always request all remote ids so we have a chance to update existing series
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} tags...", 25 + percent)

            # Request all remote tags
            response = self.do_chunked_request(
                api.get_tags_with_ids, remote_ids, callback)
            if response:
                count = self.save_tags(response)

        return count
    
    def sync_series(self, remote_ids):
        instant = datetime.utcnow()

        # @@TODO
        # local_count = (Series.select(fn.Count(Series.id))
        #                .where((Series.id << remote_ids))
        #                .scalar())
        # missing_count = len(remote_ids) - local_count
        count, missing_count = 0, 0

        # Always request all remote ids so we have a chance to update existing series
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} series...", 25 + percent)

            self.app.logger.debug(
                f"Found missing {missing_count} of {len(remote_ids)} series")
            # Request old and new series
            response = self.do_chunked_request(
                api.get_series_with_ids, remote_ids, callback)
            if response:
                count = self.save_series(response, instant)

            #  Series tags
            response = self.do_chunked_request(
                api.get_series_tags_for_ids, remote_ids, callback)
            if response:
                self.save_series_tags(response)

        return count

    def save_series(self, response, instant):
        """
        Insert new series and attempt to update existing ones
        """
        count = 0
        self.app.logger.debug("Saving series to database...")
        for batch in chunked(response, INSERT_CHUNK_SIZE):
            count += (Series.insert_many(batch)
                      .on_conflict(
                conflict_target=[Series.id],
                # Pass down values from insert clause
                preserve=[Series.imdb_id, Series.name, Series.sort_name, Series.tagline, Series.overview, Series.network,
                          Series.vote_average, Series.popularity, Series.poster_url, Series.fanart_url],
                update={Series.last_updated_on: instant})
                .as_rowcount()
                .execute())
            for series in batch:
                content = ' '.join([series['name'], series['network'], series['overview']]) 
                # FTS5 insert_many cannot handle upserts
                (SeriesIndex.insert({
                    SeriesIndex.rowid: series['id'],                    
                    SeriesIndex.name: series['name'],
                    SeriesIndex.content: content,
                })
                    # Just replace name and content edits
                    .on_conflict_replace()
                    .execute())
        SeriesIndex.optimize()
        return count

    def save_series_tags(self, response):
        self.app.logger.debug("Saving series tags to database...")
        for batch in chunked(response, INSERT_CHUNK_SIZE):
            (SeriesTag.insert_many(batch)
                # Ignore duplicate rows
                .on_conflict_ignore()
                .execute())

    def sync_episodes(self, remote_ids):
        instant = datetime.utcnow()

        # local_ids = [e.id for e in Episode.select(Episode.id)]
        # new_ids = set(remote_ids) - set(local_ids)
        count, missing_count = 0, 0

        # Always request all remote ids so we have a chance to update existing episodes
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} episodes...", 50 + percent)

            self.app.logger.debug(
                f"Found missing {missing_count} of {len(remote_ids)} episodes")
            # Request old and new episodes
            response = self.do_chunked_request(
                api.get_episodes_with_ids, remote_ids, callback)
            if response:
                count = self.save_episodes(response, instant)

        return count

    def save_episodes(self, response, instant):
        """
        Insert new episodes and attempt to update existing ones
        """
        count = 0
        self.app.logger.debug("Saving episodes to database...")
        for batch in chunked(response, INSERT_CHUNK_SIZE):
            # We need to cope with the unique constraint for (series, season, number)
            #   index because we cannot rely on episodes id's,
            #   they are often changed when TVDB users update them
            count += (Episode
                      .insert_many(batch)
                      .on_conflict(
                          conflict_target=[
                              Episode.series, Episode.season, Episode.number],
                          # Pass down values from insert clause
                          preserve=[Episode.name, Episode.overview,
                                    Episode.aired_on, Episode.thumbnail_url],
                          update={Episode.last_updated_on: instant})
                      .as_rowcount()
                      .execute())
            # EpisodeIndex.insert({
            #     EpisodeIndex.rowid: episode_id,
            #     EpisodeIndex.name: episode.name,
            #     EpisodeIndex.overview: episode.overview}).execute()
        #EpisodeIndex.optimize()            
        return count

    def sync_releases(self, remote_ids):
        instant = datetime.utcnow()

        # local_ids = [r.id for r in Release.select(Release.id)]
        # new_ids = set(remote_ids) - set(local_ids)
        count, missing_count = 0, 0

        # Always request all remote ids so we have a chance to update existing releases
        if remote_ids:
            def callback(percent, remaining):
                self.progress_callback(
                    f"Updating {remaining} torrents...", 75 + percent)

            self.app.logger.debug(
                f"Found missing {missing_count} of {len(remote_ids)} releases")
            # Request old and new releases
            response = self.do_chunked_request(
                api.get_releases_with_ids, remote_ids, callback)
            if response:
                count = self.save_releases(response, instant)

        return count

    def save_releases(self, response, instant):
        """
        Insert new releases and attempt to update existing ones
        """
        count = 0
        self.app.logger.debug("Saving releases to database...")
        for batch in chunked(response, INSERT_CHUNK_SIZE):
            count += (Release
                      .insert_many(batch)
                      .on_conflict(
                          conflict_target=[Release.id],
                          # Pass down values from insert clause
                          preserve=[Release.leechers, Release.seeders, Release.completed, Release.last_updated_on])
                      .as_rowcount()
                      .execute())
        return count

    def progress(self, value, min, max):
        return self.scale_between(value, 0, 25, min, max)

    def scale_between(self, value, min_allowed, max_allowed, min, max):
        return (max_allowed - min_allowed) * (value - min) / (max - min) + min_allowed

    def update_log(self, log, status, description):
        log.status = status
        log.description = description
        log.save()

    def do_chunked_request(self, handler, ids, callback=None):
        result = []
        ids_count = len(ids)
        for index, chunked_ids in enumerate(chunked(ids, REQUEST_CHUNK_SIZE)):
            if callback:
                percent = self.progress(index*REQUEST_CHUNK_SIZE, 0, ids_count)
                callback(percent, ids_count -
                         index*REQUEST_CHUNK_SIZE)
            self.app.logger.debug(
                f"Requesting {index + 1} of {ids_count // REQUEST_CHUNK_SIZE + 1} chunks")
            json = self.do_json_request(
                lambda: handler(self.client_id, chunked_ids))
            result.extend(json)
        return result

    def do_json_request(self, handler, retries=1):
        for index in reversed(range(retries)):
            try:
                response = handler()
                response.raise_for_status()  # Raise an exeption on HTTP errors
            except ReadTimeout as ex:
                message = f'Server timed out while handling the request, {"retrying" if index else "skipped"}'
                self.app.logger.warn(message)
                time.sleep(TIMEOUT_BEFORE_RETRY)
                continue  # Next retry
            except HTTPError as ex:
                message = f'Server error {ex.response.status_code} occurred while handling the request, giving up'
                self.app.logger.error(message)
                raise SyncError(message)
            return response.json()
        # No more retries, giving up
        raise SyncError(
            "Server timed out while handling the request. Please try again later.")



from datetime import datetime, date, timedelta

def plural(prefix, value):
    return f"{prefix}{'s' if value > 1 else ''}"


def sanitize_query(query):
    # https://www.sqlite.org/fts5.html
    sanitized_query = ""
    for c in query:
        if c.isalpha() or c.isdigit() or ord(c) > 127 or ord(c) == 96 or ord(c) == 26:
            # Allowed in FTS queries
            sanitized_query += c
        else:
            sanitized_query += " "
    return sanitized_query


def datetime_since(value, comparison_value, default="just now"):
    """
    Returns string representing "time since" e.g.
    3 days ago, 5 hours ago etc.
    """

    diff = comparison_value - value

    periods = (
        (diff.days // 365, "year", "years"),
        (diff.days // 30, "month", "months"),
        (diff.days // 7, "week", "weeks"),
        (diff.days, "day", "days"),
        (diff.seconds // 3600, "hour", "hours"),
        (diff.seconds // 60, "minute", "minutes"),
        (diff.seconds, "second", "seconds"),
    )

    for period, singular, plural in periods:
        if period:
            return "%d %s ago" % (period, singular if period == 1 else plural)

    return default



"""
Videobox package.
"""

__version__ = "0.5.0"

import os
from pathlib import Path
import click
from flask import Flask
import sqlite3
import waitress
import uuid
import videobox.models as models
import videobox.filters as filters
from .main import bp as main_blueprint
import tomli_w
try:
    import tomllib as toml  # Python 3.11+
except ImportError:
    import tomli as toml

DATABASE_FILENAME = 'library.db'
CONFIG_FILENAME = 'config.toml'
APP_DIR = Path.home().joinpath(".videobox")


def create_app(config_class=None):
    app = Flask(__name__)
    os.makedirs(APP_DIR, exist_ok=True)

    if config_class:    
        app.config.from_object(config_class)
    else:
        app.config['DATABASE_URL'] = f'sqlite:///{APP_DIR.joinpath(DATABASE_FILENAME)}' 
        config_path = os.path.join(APP_DIR, CONFIG_FILENAME)
        if app.config.from_file(config_path, load=toml.load, text=False, silent=True):
            app.logger.debug(f"Using configuration file found in {APP_DIR}")
        else:
            config = get_default_config()
            app.config.from_mapping(config)
            with open(config_path, "wb") as f:
                tomli_w.dump(config, f)

    # Initialize Flask extensions here
    models.db_wrapper.init_app(app)
    models.db_wrapper.database.pragma('foreign_keys', 1, permanent=True)
    models.db_wrapper.database.pragma('journal_mode', 'wal', permanent=True)
    app.logger.debug(f"Using SQLite version {sqlite3.sqlite_version}")

    # Make sure db schema is updated 
    models.setup()
    app.logger.debug(f"Using database found in {APP_DIR}")

    # Register main app routes
    app.register_blueprint(main_blueprint)

    # Register custom template filters
    filters.init_app(app)

    return app

def get_default_config():
    return {"API_CLIENT_ID": uuid.uuid1().hex}

@click.command()
@click.option('--host', help='Hostname or IP address on which to listen, default is 0.0.0.0, which means "all IP addresses on this host".', default="0.0.0.0")
@click.option('--port', help='TCP port on which to listen, default is 8080', type=int, default=8080)
def serve(host, port):
    print(f'Server started. Point your browser to http://{host}:{port} to use the web interface.')
    waitress.serve(create_app(), host=host, port=port, threads=8)